package org.processmining.markovprocess.models;

import static org.processmining.plugins.stationarydistribution.XesAuxiliary.getTime;
import static org.processmining.plugins.stationarydistribution.XesAuxiliary.id;
import static org.processmining.plugins.stationarydistribution.XesAuxiliary.isStart;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;

public class ContinuousTimeMarkovChain  implements Map<String, Double>{
	private Map<String,List<Double>> transitionTimes;
	
	// Keep dtmc for jump probabilities
	private DiscreteTimeMarkovChain dtmc;
	private Map<String, Double> rates;
	private Map<String, List<Double>> lifetimes;
	
	private double rateUpperBound = 0;
	
	public ContinuousTimeMarkovChain(XLog log) {
		// Get the discrete chain
		dtmc = new DiscreteTimeMarkovChain(log);
		
		// Remove transitions to self
		dtmc = dtmc.removeTransitionsToSelf();
		
		transitionTimes = new HashMap<>();
		lifetimes = new HashMap<>();
		
		// Add the states to the maps
		for (String state : dtmc.keySet()) {
			transitionTimes.put(state, new ArrayList<Double>());
			lifetimes.put(state, new ArrayList<Double>());
		}
		
		// Find big negative transition times caused by transitions from winter to summer time
		List<XTrace> illegalTraces = new ArrayList<XTrace>();
		for (XTrace trace : log) {
			
			Map<String,List<Double>> arrivalTimes = new HashMap<>();
			
			String state = null;
			XEvent lastEvent = null;
			Double time = null;
			for (XEvent event : trace) {
				// First transition
				if (time == null) {
					time = getTime(event);
					state = id(event);
				}
				
				// New transition
				if(!id(event).equals(state)) {
					if (getTime(lastEvent)-time<-1) {
						illegalTraces.add(trace);
					}
					time = getTime(event);
					state = id(event);
				}
				lastEvent = event;
				
				// Add arrival time to list
				if (isStart(event)) {
					if (!arrivalTimes.containsKey(state)) {
						arrivalTimes.put(state, new ArrayList<Double>());
					}
					
					arrivalTimes.get(state).add(time);
				}
			}
			
			// Add new lifetimes to the map
			for (Entry<String,List<Double>> entry : arrivalTimes.entrySet()) {
				for (Double arrivalTime : entry.getValue()) {
					lifetimes.get(entry.getKey()).add(time-arrivalTime);
				}
			}
		}
		
		// Remove those traces
		for (XTrace illegalTrace : illegalTraces)
			log.remove(illegalTrace);
		
		// Gather the times the ctmc stayed in each state.
		for (XTrace trace : log) {
			String state = null;
			XEvent lastEvent = null;
			Double time = null;
			for (XEvent event : trace) {
				// First transition
				if (time == null) {
					time = getTime(event);
					state = id(event);
				}
				
				// New transition
				if(!id(event).equals(state)) {
					double transitionTime = getTime(lastEvent)-time;
					
					transitionTimes.get(state).add(transitionTime > 0 ? transitionTime : Double.MIN_NORMAL);
					time = getTime(event);
					state = id(event);
				}
				lastEvent = event;
			}
			// Last transition of the trace
			double transitionTime = getTime(lastEvent)-time;
			transitionTimes.get(state).add(transitionTime > 0 ? transitionTime : Double.MIN_NORMAL);
		}
		
		// Find total transition intensity
		rates = new HashMap<String,Double>();
		for (Entry<String, List<Double>> entry : transitionTimes.entrySet()) {
			double sum = 0;
			for (Double time : entry.getValue()) {
				sum += time;
			}
			rates.put(entry.getKey(), sum > 0 ? entry.getValue().size()/sum : 0);
			if (rates.get(entry.getKey()) > rateUpperBound) {
				rateUpperBound = rates.get(entry.getKey());
			}
		}
	}
	
	public StateLifetimes getStateLifetimes() {
		return new StateLifetimes(lifetimes);
	}
	
	public double getIntensity(String from, String to) {
		if (!from.equals(to)) {
			return dtmc.get(from).containsKey(to) ? rates.get(from)*dtmc.get(from).get(to) : 0;
		} else {
			return -rates.get(from);
		}
	}
	
	public double getJumpProbability(String from, String to) {
		return dtmc.get(from).get(to);
	}
	
	public Map<String,List<Double>> getTransitionTimes() {
		return transitionTimes;
	}
	
	public List<String> getIncomingStates(String to) {
		return dtmc.getIncomingStates(to);
	}

	public List<String> getOutgoingStates(String from) {
		return dtmc.getOutgoingStates(from);
	}
	
	public DiscreteTimeMarkovChain getDtmc() {
		return dtmc;
	}
	
	public double getRateUpperBound() {
		return rateUpperBound;
	}

	public int size() {
		return rates.size();
	}

	public boolean isEmpty() {
		return rates.isEmpty();
	}

	public boolean containsKey(Object key) {
		return rates.containsKey(key);
	}

	public boolean containsValue(Object value) {
		return rates.containsValue(value);
	}

	public Double get(Object key) {
		return rates.get(key);
	}

	public Double put(String key, Double value) {
		if (value > rateUpperBound)
			rateUpperBound = value;
		return rates.put(key, value);
	}

	public Double remove(Object key) {
		return rates.remove(key);
	}

	public void putAll(Map<? extends String, ? extends Double> m) {
		rates.putAll(m);
	}

	public void clear() {
		rates.clear();
	}

	public Set<String> keySet() {
		return rates.keySet();
	}

	public Collection<Double> values() {
		return rates.values();
	}

	public Set<java.util.Map.Entry<String, Double>> entrySet() {
		return rates.entrySet();
	}
}
