package org.processmining.markovprocess.models;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.stationarydistribution.XesAuxiliary;

/**
 * This class encapsulates Map<String,Map<String,Double>>. It is created to be able to use this as input for other plugins
 * and to group auxiliary functions.
 * @author s147569
 */
public class DiscreteTimeMarkovChain implements Map<String,Map<String,Double>> {
	// The graph: Each 'state' is mapped to a mapping from outgoing states to the number
	// of times it occurred
	private Map<String, Map<String,Double>> dtmc = new HashMap<String,Map<String,Double>>();
	private InitialDistribution initialDistribution = new InitialDistribution();
	
	public static final String failedState = "F";
	
	// The keys/states, can be reordered for clustering purposes.
	private List<String> states = new ArrayList<String>();
	
	/**
	 * Returns the state in their current ordering.
	 * @return The states in their current order.
	 */
	public List<String> getStatesInOrder() {
		return states;
	}
	
	public void setStatesOrder(List<String> order) {
		states = order;
	}
	
	/*
	 * This constructor creates the DTMC from the log
	 */
	public DiscreteTimeMarkovChain(XLog log) {
		// Add a failed state
		states.add(failedState);
		dtmc.put(failedState, new HashMap<String,Double>());
		dtmc.get(failedState).put(failedState, 1d);
		initialDistribution.put(failedState, 0d);
		
		for (XTrace trace : log) {
			String previous = null;
			for (XEvent event : trace) {
				if (XesAuxiliary.isStart(event)) {
					String eventID = XesAuxiliary.id(event);
					if (!states.contains(eventID)) {
						states.add(eventID);
						dtmc.put(eventID, new HashMap<String,Double>());
					}
	
					
					if (previous == null) {
						// First state
						initialDistribution.put(eventID, 
								(initialDistribution.containsKey(eventID) 
										? initialDistribution.get(eventID) : 0d)
								+1d/log.size());
					} else {
						// If it is not the first event in the trace
						addTransition(previous, eventID);
					}
					previous = eventID;
				}
			}
			// Add transition to failed (null-check for empty traces)
			if (previous != null) {
				addTransition(previous, failedState);
			}
		}

		// normalize:
		for (String state : states) {
			double sum = 0;
			for (double weight : dtmc.get(state).values())
				sum+=weight;
			for (String transition : dtmc.get(state).keySet())
				dtmc.get(state).put(transition, dtmc.get(state).get(transition)/sum);
		}
	}
	
	private DiscreteTimeMarkovChain(Collection<String> states) {
		dtmc = new HashMap<String,Map<String,Double>>();
		this.states = new ArrayList<String>();
		for (String state : states) {
			this.states.add(state);
			dtmc.put(state, new HashMap<String,Double>());
		}
	}
	
	// Increments the number of occurrences from previous to next
	private void addTransition(String previous, String next) {

		Map<String, Double> outgoing = dtmc.get(previous);

		// Make sure there is an 'edge' for this transition
		if (!outgoing.containsKey(next))
			outgoing.put(next, 0d);

		// Increment
		outgoing.put(next, outgoing.get(next)+1);
	}
	
	public DiscreteTimeMarkovChain removeTransitionsToSelf() {
		DiscreteTimeMarkovChain newDtmc = new DiscreteTimeMarkovChain(getStatesInOrder());
		// Remove transitions to self
		for (String state : newDtmc.keySet()) {
			// We keep transitions to self for the failedState
			if (state.equals(failedState)) {
				newDtmc.get(failedState).put(failedState, 1d);
			} else {
				double divider = 1;
				if (dtmc.get(state).containsKey(state)) {
					divider = 1-dtmc.get(state).get(state);
				}
				// Make sure the outgoing probabilities sum up to 1
				for (String out : dtmc.get(state).keySet())
					if (!out.equals(state))
						newDtmc.get(state).put(out, dtmc.get(state).get(out)/divider);
			}
		}
		
		// Copy initial distribution
		newDtmc.getInitialDistribution().putAll(getInitialDistribution());
		return newDtmc;
	}
	
	public static DiscreteTimeMarkovChain uniformize(ContinuousTimeMarkovChain ctmc) {
		DiscreteTimeMarkovChain newDtmc = new DiscreteTimeMarkovChain(ctmc.getDtmc().getStatesInOrder());
		for (String from : ctmc.getDtmc().getStatesInOrder()) {
			if (from.equals(failedState)) {
				// Do not normalize the failedState
				newDtmc.put(from, ctmc.getDtmc().get(from));
			} else {
				for (String to : ctmc.getDtmc().getStatesInOrder()) {
					// Page 247 from Bertsekas volume 2
					if (from.equals(to)) {
						if (!ctmc.getDtmc().get(from).containsKey(to))
							newDtmc.get(from).put(to, -ctmc.get(from)/ctmc.getRateUpperBound()+1);
						else
							newDtmc.get(from).put(to, ctmc.get(from)*(ctmc.getDtmc().get(from).get(to)-1)/ctmc.getRateUpperBound()+1);
					} else
						if (!ctmc.getDtmc().get(from).containsKey(to))
							newDtmc.get(from).put(to, 0d);
						else
							newDtmc.get(from).put(to, ctmc.get(from)*ctmc.getDtmc().get(from).get(to)/ctmc.getRateUpperBound());
				}
			}
		}
		
		// Copy initial distribution
		newDtmc.getInitialDistribution().putAll(ctmc.getDtmc().getInitialDistribution());
		return newDtmc;
	}
	
	public DiscreteTimeMarkovChain removeFailedState() {
		DiscreteTimeMarkovChain newDtmc = new DiscreteTimeMarkovChain(getStatesInOrder());
		newDtmc.remove(failedState);
		for (String from : newDtmc.getStatesInOrder())
			for (String to : dtmc.get(from).keySet())
				if (to != failedState)
					newDtmc.get(from).put(to, dtmc.get(from).get(to));
		
		// Copy initial distribution
		newDtmc.getInitialDistribution().putAll(getInitialDistribution());
		return newDtmc;
	}
	
	public List<String> getOutgoingStates(String from) {
		List<String> outgoing = new ArrayList<>();
		for (Entry<String,Double> entry : dtmc.get(from).entrySet()) {
			if (entry.getValue() > 0 && !entry.getKey().equals(from))
				outgoing.add(entry.getKey());
		}
		return outgoing;
	}
	
	public List<String> getIncomingStates(String to) {
		List<String> incoming = new ArrayList<>();
		for (String from : states) {
			if (!from.equals(to) && this.getTransitionProbability(from, to) > 0) {
				incoming.add(from);
			}
		}
		return incoming;
	}
	
	public StateClustering getCluster(int clusterDepth) {
		StationaryDistribution sd = new StationaryDistribution(this);
		return new StateClustering(sd,clusterDepth);
	}
	
	public double getTransitionProbability(String from, String to) {
		 return dtmc.get(from).containsKey(to) ? dtmc.get(from).get(to) : 0;
	}
	
	public Map<String,Double> getInitialDistribution() {
		return initialDistribution;
	}
	
	/* 
	 * Implementation of inherited methods from Map<String,Map<String,Double>>.
	 */
	@Override
	public int size() {
		return dtmc.size();
	}

	@Override
	public boolean isEmpty() {
		return dtmc.isEmpty();
	}

	@Override
	public boolean containsKey(Object key) {
		return dtmc.containsKey(key);
	}

	@Override
	public boolean containsValue(Object value) {
		return dtmc.containsValue(value);
	}

	@Override
	public Map<String, Double> get(Object key) {
		return dtmc.get(key);
	}

	@Override
	public Map<String, Double> put(String key, Map<String, Double> value) {
		return dtmc.put(key, value);
	}

	@Override
	public Map<String, Double> remove(Object key) {
		return dtmc.remove(key);
	}

	@Override
	public void putAll(Map<? extends String, ? extends Map<String, Double>> m) {
		dtmc.putAll(m);
	}

	@Override
	public void clear() {
		dtmc.clear();
	}

	@Override
	public Set<String> keySet() {
		return dtmc.keySet();
	}

	@Override
	public Collection<Map<String, Double>> values() {
		return dtmc.values();
	}

	@Override
	public Set<Entry<String, Map<String, Double>>> entrySet() {
		return dtmc.entrySet();
	}

}
